"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

The class ICMio provides read access to S5p Tropomi ICM_CA_SIR products

Copyright (c) 2017-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta
from pathlib import Path, PurePosixPath
from setuptools_scm import get_version

import h5py
import numpy as np


# - global parameters ------------------------------

# - local functions --------------------------------

# - class definition -------------------------------
class ICMio():
    """
    This class should offer all the necessary functionality to read Tropomi
    ICM_CA_SIR products

    Attributes
    ----------
    fid : h5py.File
    filename : string
    bands : string

    Methods
    -------
    coverage_time
       Returns start and end of the measurement coverage time.
    creation_time
       Returns creation date of this product
    orbit
       Returns value of revolution counter.
    processor_version
       Returns version of the L01b processor used to generate this product.
    close()
       Close resources
    find(msm_class)
       Find a measurement as <processing-class name>.
    select(msm_type: str, msm_path=None)
       Select a measurement as <processing class>_<ic_id>.
    get_attr(attr_name)
       Obtain value of an HDF5 file attribute.
    get_ref_time(band=None)
       Returns reference start time of measurements.
    get_delta_time(band=None)
       Returns offset from the reference start time of measurement.
    get_instrument_settings(band=None)
       Returns instrument settings of measurement.
    get_exposure_time(band=None)
       Returns pixel exposure time of the measurements, which is calculated
       from the parameters 'int_delay' and 'int_hold' for SWIR.
    get_housekeeping_data(band=None)
       Returns housekeeping data of measurements.
    get_msmt_keys(band=None)
       Read msmt_keys from the analysis groups.
    get_msm_attr(msm_dset, attr_name, band=None)
       Returns attribute of measurement dataset 'msm_dset'.
    get_geo_data(band=None, geo_dset='satellite_latitude,satellite_longitude')
       Returns data of selected datasets from the GEODATA group.
    get_msm_data(msm_dset, band='78', *, read_raw=False, columns=None,
                 fill_as_nan=True)
       Read datasets from a measurement selected by class-method 'select'
    read_direct_msm(msm_dset, dest_sel=None, dest_dtype=None, fill_as_nan=False)
       The faster implementation of class method 'get_msm_data'.
    set_housekeeping_data(data, band=None)
       Returns housekeeping data of measurements.
    set_msm_data(msm_dset, data, band='78')
       Alter dataset from a measurement selected using function 'select'.

    Notes
    -----

    Examples
    --------
    """
    def __init__(self, icm_product, readwrite=False):
        """
        Initialize access to an ICM product

        Parameters
        ----------
        icm_product :  string
           full path to in-flight calibration measurement product
        readwrite   :  boolean
           open product in read-write mode (default is False)
        """
        if not Path(icm_product).is_file():
            raise FileNotFoundError('{} does not exist'.format(icm_product))

        # initialize class-attributes
        self.__rw = readwrite
        self.__msm_path = None
        self.__patched_msm = []
        self.filename = icm_product
        self.bands = None

        # open ICM product as HDF5 file
        if readwrite:
            self.fid = h5py.File(icm_product, "r+")
        else:
            self.fid = h5py.File(icm_product, "r")

    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r}, readwrite={!r})'.format(class_name,
                                                 self.filename, self.__rw)

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    # def __del__(self):
    #    """
    #    called when the object is destroyed
    #    """
    #    self.close()

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self):
        """
        Before closing the product, we make sure that the output product
        describes what has been altered by the S/W. To keep any change
        traceable.

        as attributes of this group, we write:
         - dateStamp ('now')
         - Git-version of S/W
         - list of patched datasets
         - auxiliary datasets used by patch-routines
        """
        if self.fid is None:
            return

        self.bands = None
        if self.__patched_msm:
            sgrp = self.fid.require_group("METADATA/SRON_METADATA")
            sgrp.attrs['dateStamp'] = datetime.utcnow().isoformat()
            sgrp.attrs['git_tag'] = get_version(root='..',
                                                relative_to=__file__)
            if 'patched_datasets' not in sgrp:
                dtype = h5py.special_dtype(vlen=str)
                dset = sgrp.create_dataset('patched_datasets',
                                           (len(self.__patched_msm),),
                                           maxshape=(None,), dtype=dtype)
                dset[:] = np.asarray(self.__patched_msm)
            else:
                dset = sgrp['patched_datasets']
                dset.resize(dset.shape[0] + len(self.__patched_msm), axis=0)
                dset[dset.shape[0]-1:] = np.asarray(self.__patched_msm)

        self.fid.close()
        self.fid = None

    # ---------- RETURN VERSION of the S/W ----------
    def find(self, msm_class) -> list:
        """
        find a measurement as <processing-class name>

        Parameters
        ----------
        msm_class :  string
          processing-class name without ICID

        Returns
        -------
        out  :  list of strings
           String with msm_type as used by ICMio.select
        """
        res = []

        grp_list = ['ANALYSIS', 'CALIBRATION', 'IRRADIANCE', 'RADIANCE']
        for ii in '12345678':
            for name in grp_list:
                grp_name = 'BAND{}_{}'.format(ii, name)
                if grp_name in self.fid:
                    gid = self.fid[grp_name]
                    res += [s for s in gid if s.startswith(msm_class)]

        return list(set(res))

    # -------------------------
    def select(self, msm_type: str, msm_path=None) -> str:
        """
        Select a measurement as <processing class>_<ic_id>

        Parameters
        ----------
        msm_type :  string
           Name of measurement group
        msm_path : {'BAND%_ANALYSIS', 'BAND%_CALIBRATION',
                    'BAND%_IRRADIANCE', 'BAND%_RADIANCE'}
           Name of path in HDF5 file to measurement group

        Returns
        -------
        string
           String with spectral bands found in product or empty

        Attributes
        ----------
        bands : string
           Available spectral bands (or empty)
        __msm_path : string
           Full name of selected group in file (or None)
        """
        self.bands = ''
        self.__msm_path = None

        # if path is given, then only determine avaialble spectral bands
        # else determine path and avaialble spectral bands
        if msm_path is None:
            grp_list = ['ANALYSIS', 'CALIBRATION', 'IRRADIANCE', 'RADIANCE']
            for ii in '12345678':
                for name in grp_list:
                    grp_path = PurePosixPath('BAND{}_{}'.format(ii, name),
                                             msm_type)
                    if str(grp_path) in self.fid:
                        msm_path = 'BAND{}_{}'.format('%', name)
                        self.bands += ii
        else:
            if not msm_path.startswith('BAND%'):
                raise ValueError('msm_path should start with BAND%')

            for ii in '12345678':
                grp_path = PurePosixPath(msm_path.replace('%', ii), msm_type)
                if str(grp_path) in self.fid:
                    self.bands += ii

        # return in case no data was found
        if self.bands:
            self.__msm_path = PurePosixPath(msm_path, msm_type)

        return self.bands

    # ---------- Functions that work before MSM selection ----------
    @property
    def orbit(self) -> int:
        """
        Returns reference orbit number
        """
        if 'reference_orbit' in self.fid.attrs:
            return int(self.fid.attrs['reference_orbit'])

        return None

    @property
    def processor_version(self) -> str:
        """
        Returns version of the L01b processor
        """
        if 'processor_version' not in self.fid.attrs:
            return None

        res = self.fid.attrs['processor_version']
        if isinstance(res, bytes):
            # pylint: disable=no-member
            return res.decode('ascii')

        return res

    @property
    def coverage_time(self) -> tuple:
        """
        Returns start and end of the measurement coverage time
        """
        if 'time_coverage_start' not in self.fid.attrs \
           or 'time_coverage_end' not in self.fid.attrs:
            return None

        res1 = self.fid.attrs['time_coverage_start']
        if isinstance(res1, bytes):
            # pylint: disable=no-member
            res1 = res1.decode('ascii')

        res2 = self.fid.attrs['time_coverage_end']
        if isinstance(res2, bytes):
            # pylint: disable=no-member
            res2 = res2.decode('ascii')

        return (res1, res2)

    @property
    def creation_time(self) -> str:
        """
        Returns version of the L01b processor
        """
        grp = self.fid['/METADATA/ESA_METADATA/earth_explorer_header']
        dset = grp['fixed_header/source']
        return dset.attrs['Creation_Date'].split(b'=')[1].decode('ascii')

    def get_attr(self, attr_name):
        """
        Obtain value of an HDF5 file attribute

        Parameters
        ----------
        attr_name : string
           name of the attribute
        """
        if attr_name not in self.fid.attrs:
            return None

        res = self.fid.attrs[attr_name]
        if isinstance(res, bytes):
            return res.decode('ascii')

        return res

    # ---------- Functions that only work after MSM selection ----------
    def get_ref_time(self, band=None) -> datetime:
        """
        Returns reference start time of measurements

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        ref_time = datetime(2010, 1, 1, 0, 0, 0)
        if not self.__msm_path:
            return ref_time

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            grp = self.fid[msm_path]
            dset = grp[msm_type.lower() + '_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'OBSERVATIONS')
                grp = self.fid[str(grp_path)]
                ref_time += timedelta(seconds=int(grp['time'][0]))
        elif msm_type in ['DPQF_MAP', 'NOISE']:
            grp_path = PurePosixPath(msm_path).parent / 'ANALOG_OFFSET_SWIR'
            grp = self.fid[str(grp_path)]
            dset = grp['analog_offset_swir_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'OBSERVATIONS')
                grp = self.fid[str(grp_path)]
                ref_time += timedelta(seconds=int(grp['time'][0]))
        else:
            grp_path = PurePosixPath(msm_path, 'OBSERVATIONS')
            grp = self.fid[str(grp_path)]
            ref_time += timedelta(seconds=int(grp['time'][0]))

        return ref_time

    def get_delta_time(self, band=None):
        """
        Returns offset from the reference start time of measurement

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if not self.__msm_path:
            return None

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        res = None
        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            grp = self.fid[msm_path]
            dset = grp[msm_type.lower() + '_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'OBSERVATIONS')
                grp = self.fid[str(grp_path)]
                if res is None:
                    res = grp['delta_time'][0, :].astype(int)
                else:
                    res = np.append(res, grp['delta_time'][0, :].astype(int))
        elif msm_type in ['DPQF_MAP', 'NOISE']:
            grp_path = PurePosixPath(msm_path).parent / 'ANALOG_OFFSET_SWIR'
            grp = self.fid[str(grp_path)]
            dset = grp['analog_offset_swir_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'OBSERVATIONS')
                grp = self.fid[grp_path]
                if res is None:
                    res = grp['delta_time'][0, :].astype(int)
                else:
                    res = np.append(res, grp['delta_time'][0, :].astype(int))
        else:
            grp_path = PurePosixPath(msm_path, 'OBSERVATIONS')
            grp = self.fid[str(grp_path)]
            res = grp['delta_time'][0, :].astype(int)

        return res

    def get_instrument_settings(self, band=None):
        """
        Returns instrument settings of measurement

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if not self.__msm_path:
            return None

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        res = None
        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            grp = self.fid[msm_path]
            dset = grp[msm_type.lower() + '_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'INSTRUMENT')
                grp = self.fid[str(grp_path)]
                if res is None:
                    res = grp['instrument_settings'][:]
                else:
                    res = np.append(res, grp['instrument_settings'][:])
        elif msm_type == 'DPQF_MAP':
            grp_path = PurePosixPath(msm_path).parent / 'ANALOG_OFFSET_SWIR'
            grp = self.fid[str(grp_path)]
            dset = grp['analog_offset_swir_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'INSTRUMENT')
                grp = self.fid[grp_path]
                if res is None:
                    res = grp['instrument_settings'][:]
                else:
                    res = np.append(res, grp['instrument_settings'][:])
        elif msm_type == 'NOISE':
            grp = self.fid[msm_path]
            dset = grp[msm_type.lower() + '_msmt_keys']
            icid = dset['icid'][dset.size // 2]
            grp_path = PurePosixPath(
                'BAND{}_CALIBRATION'.format(band),
                'BACKGROUND_RADIANCE_MODE_{:04d}'.format(icid),
                'INSTRUMENT')
            grp = self.fid[str(grp_path)]
            res = grp['instrument_settings'][:]
        else:
            grp_path = PurePosixPath(msm_path, 'INSTRUMENT')
            grp = self.fid[str(grp_path)]
            res = grp['instrument_settings'][:]

        return res

    def get_exposure_time(self, band=None) -> list:
        """
        Returns pixel exposure time of the measurements, which is calculated
        from the parameters 'int_delay' and 'int_hold' for SWIR.

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band
        """
        # obtain instrument settings
        instr_arr = self.get_instrument_settings(band)
        if instr_arr is None:
            return None

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        # calculate exact exposure time
        res = []
        for instr in instr_arr:
            if int(band) > 6:
                res.append(1.25e-6 * (65540
                                      - instr['int_delay'] + instr['int_hold']))
            else:
                res.append(instr['exposure_time'])

        return res

    def get_housekeeping_data(self, band=None):
        """
        Returns housekeeping data of measurements

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if not self.__msm_path:
            return None

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        res = None
        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            grp = self.fid[msm_path]
            dset = grp[msm_type.lower() + '_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'INSTRUMENT')
                grp = self.fid[str(grp_path)]
                if res is None:
                    res = np.squeeze(grp['housekeeping_data'])
                else:
                    res = np.append(res, np.squeeze(grp['housekeeping_data']))
        elif msm_type in ['DPQF_MAP', 'NOISE']:
            grp_path = PurePosixPath(msm_path).parent / 'ANALOG_OFFSET_SWIR'
            grp = self.fid[str(grp_path)]
            dset = grp['analog_offset_swir_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'INSTRUMENT')
                grp = self.fid[str(grp_path)]
                if res is None:
                    res = np.squeeze(grp['housekeeping_data'])
                else:
                    res = np.append(res, np.squeeze(grp['housekeeping_data']))
        else:
            grp_path = PurePosixPath(msm_path, 'INSTRUMENT')
            grp = self.fid[str(grp_path)]
            res = np.squeeze(grp['housekeeping_data'])

        return res

    # -------------------------
    def get_msmt_keys(self, band=None):
        """
        Read msmt_keys from the analysis groups

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band

        Returns
        -------
         [ANALOG_OFFSET_SWIR] analog_offset_swir_group_keys
         [LONG_TERM_SWIR]     long_term_swir_group_keys
         [NOISE]              noise_msmt_keys
         else                 None
        """
        if not self.__msm_path:
            return None

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            grp = self.fid[msm_path]
            return np.squeeze(grp[msm_type.lower() + '_group_keys'])

        if msm_type == 'NOISE':
            grp = self.fid[msm_path]
            return np.squeeze(grp[msm_type.lower() + '_msmt_keys'])

        return None

    # -------------------------
    def get_msm_attr(self, msm_dset, attr_name, band=None):
        """
        Returns attribute of measurement dataset "msm_dset"

        Parameters
        ----------
        msm_dset  :  string
            Name of measurement dataset
        attr_name :  string
            Name of the attribute
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band

        Returns
        -------
        out   :   scalar or numpy array
           value of attribute "attr_name"
        """
        if not self.__msm_path:
            return None

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        for dset_grp in ['OBSERVATIONS', 'ANALYSIS', '']:
            ds_path = PurePosixPath(str(self.__msm_path).replace('%', band),
                                    dset_grp, msm_dset)
            if str(ds_path) not in self.fid:
                continue

            if attr_name in self.fid[str(ds_path)].attrs:
                attr = self.fid[str(ds_path)].attrs[attr_name]
                if isinstance(attr, bytes):
                    return attr.decode('ascii')

                return attr

        return None

    def get_geo_data(self, band=None,
                     geo_dset='satellite_latitude,satellite_longitude'):
        """
        Returns data of selected datasets from the GEODATA group

        Parameters
        ----------
        geo_dset  :  string
            Name(s) of datasets in the GEODATA group, comma separated
            Default is 'satellite_latitude,satellite_longitude'
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band

        Returns
        -------
        out   :   dictionary
           dictionary data of selected datasets from the GEODATA group
           names of dictionary are taken from parameter geo_dset
        """
        if not self.__msm_path:
            return None

        if band is None:
            band = str(self.bands[0])
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        res = {}
        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            grp = self.fid[msm_path]
            dset = grp[msm_type.lower() + '_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'GEODATA')
                grp = self.fid[str(grp_path)]
                for key in geo_dset.split(','):
                    res[key] = np.squeeze(grp[key])
        elif msm_type in ['DPQF_MAP', 'NOISE']:
            grp_path = PurePosixPath(msm_path).parent / 'ANALOG_OFFSET_SWIR'
            grp = self.fid[str(grp_path)]
            dset = grp['analog_offset_swir_group_keys']
            group_keys = dset['group'][:]
            for name in group_keys:
                grp_path = PurePosixPath('BAND{}_CALIBRATION'.format(band),
                                         name.decode('ascii'),
                                         'GEODATA')
                grp = self.fid[str(grp_path)]
                for key in geo_dset.split(','):
                    res[key] = np.squeeze(grp[key])
        else:
            grp_path = PurePosixPath(msm_path, 'GEODATA')
            grp = self.fid[str(grp_path)]
            for key in geo_dset.split(','):
                res[key] = np.squeeze(grp[key])

        return res

    def get_msm_data(self, msm_dset, band='78', *, read_raw=False,
                     columns=None, fill_as_nan=True):
        """
        Read datasets from a measurement selected by class-method "select"

        Parameters
        ----------
        msm_dset   :  string
            name of measurement dataset
            if msm_dset is None then show names of available datasets
        band       :  {'1', '2', '3', ..., '8', '12', '34', '56', '78'}
            Select data from one spectral band or channel
            Default is '78' which combines band 7/8 to SWIR detector layout
        read_raw   : boolean
            Perform raw read: without slicing or data conversion,
            and ignore keywords: colums, fill_as_nan.
            Default: False
        columns    : [i, j]
            Slice data on fastest axis (columns) as from index 'i' to 'j'
        fill_as_nan :  boolean
            Replace (float) FillValues with Nan's, when True

        Returns
        -------
        out  :  array
           Data of measurement dataset "msm_dset"
        """
        fillvalue = float.fromhex('0x1.ep+122')

        if not self.__msm_path:
            return None

        if not isinstance(band, str):
            raise TypeError('band must be a string')

        if band not in self.bands:
            raise ValueError('band not found in product')

        data = []
        if read_raw:
            for ii in band:
                for dset_grp in ['OBSERVATIONS', 'ANALYSIS', '']:
                    msm_path = str(self.__msm_path).replace('%', ii)
                    ds_path = PurePosixPath(msm_path, dset_grp, msm_dset)
                    if str(ds_path) not in self.fid:
                        continue

                    data.append(np.squeeze(self.fid[str(ds_path)]))

            return data

        # skip row257 from the SWIR detector
        rows = None
        if int(band[0]) > 6:
            rows = [0, -1]

        # list potential names of the dataset dimensions
        time_list = ['time', 'scanline']
        row_list = ['width', 'pixel', 'pixel_window', 'ground_pixel']
        column_list = ['height', 'spectral_channel', 'spectral_channel_window']

        column_dim = None   # column dimension is unknown
        for ii in band:
            for dset_grp in ['OBSERVATIONS', 'ANALYSIS', '']:
                msm_path = str(self.__msm_path).replace('%', ii)
                ds_path = PurePosixPath(msm_path, dset_grp, msm_dset)
                if str(ds_path) not in self.fid:
                    continue
                dset = self.fid[str(ds_path)]

                skipped = 0
                data_sel = ()
                for xx in range(dset.ndim):
                    if len(dset.dims[xx][0][:]) == 1:
                        skipped += 1

                    dim_name = PurePosixPath(dset.dims[xx][0].name).name
                    if dim_name in time_list:
                        data_sel += (slice(None),)
                    elif dim_name in row_list:
                        if rows is None:
                            data_sel += (slice(None),)
                        else:
                            data_sel += (slice(*rows),)
                    elif dim_name in column_list:
                        column_dim = xx - skipped
                        if columns is None:
                            data_sel += (slice(None),)
                        else:
                            data_sel += (slice(*columns),)
                    else:
                        raise ValueError

                if dset.dtype == np.float32:
                    res = np.squeeze(dset.astype(float)[data_sel])
                else:
                    res = np.squeeze(dset[data_sel])

                if fill_as_nan and dset.attrs['_FillValue'] == fillvalue:
                    res[(res == fillvalue)] = np.nan
                data.append(res)

        # Note the current implementation will not work for channels where
        # the output of its bands can have different spatial dimensions (rows)
        # or different integration times (frames/scanlines)
        #
        # no data found
        if not data:
            return None
        # return selected band
        if len(data) == 1:
            return data[0]
        # return bands stacked
        if column_dim is None:
            return data        # np.stack(data)
        # return band in detector layout
        return np.concatenate(data, axis=column_dim)

    def read_direct_msm(self, msm_dset, dest_sel=None,
                        dest_dtype=None, fill_as_nan=False):
        """
        The faster implementation of get_msm_data()

        Parameters
        ----------
        msm_dset    :  string
            Name of measurement dataset
        dest_sel    :  numpy slice
            Selection must be the output of numpy.s_[<args>].
        dest_dtype  :  numpy dtype
            Perform type conversion
        fill_as_nan :  boolean
            Replace (float) FillValues with Nan's, when True

        Returns
        -------
        out  :  list
           list with data of all available bands
        """
        fillvalue = float.fromhex('0x1.ep+122')

        if not self.__msm_path:
            return None

        if dest_sel is None:
            dest_sel = np.s_[...]

        data = []
        for ii in self.bands:
            msm_path = str(self.__msm_path).replace('%', ii)
            for dset_grp in ['OBSERVATIONS', 'ANALYSIS', '']:
                ds_path = str(PurePosixPath(msm_path, dset_grp, msm_dset))
                if ds_path not in self.fid:
                    continue

                dset = self.fid[ds_path]
                if dest_dtype is None:
                    buff = dset[dest_sel]
                    if fill_as_nan and '_FillValue' in dset.attrs:
                        if np.issubdtype(buff.dtype, np.floating):
                            fillvalue = dset.attrs['_FillValue'][0]
                            buff[(buff == fillvalue)] = np.nan
                else:
                    buff = dset.astype(dest_dtype)[dest_sel]
                    if fill_as_nan and '_FillValue' in dset.attrs:
                        if np.issubdtype(buff.dtype, np.floating):
                            buff[(buff == fillvalue)] = np.nan

                data.append(buff)

        return data

    # -------------------------
    def set_housekeeping_data(self, data, band=None) -> None:
        """
        Returns housekeeping data of measurements

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if not self.__rw:
            raise PermissionError('read/write access required')

        if not self.__msm_path:
            return

        if band is None:
            band = self.bands[0]
        elif band not in self.bands:
            raise ValueError('band not found in product')

        msm_path = str(self.__msm_path).replace('%', band)
        msm_type = self.__msm_path.name

        if msm_type in ['ANALOG_OFFSET_SWIR', 'LONG_TERM_SWIR']:
            pass
        elif msm_type in ['DPQF_MAP', 'NOISE']:
            pass
        else:
            ds_path = PurePosixPath(msm_path, 'INSTRUMENT', 'housekeeping_data')
            self.fid[str(ds_path)][0, :] = data

            self.__patched_msm.append(str(ds_path))

    def set_msm_data(self, msm_dset, data, band='78') -> None:
        """
        Alter dataset from a measurement selected using function "select"

        Parameters
        ----------
        msm_dset   :  string
            name of measurement dataset
        band       :  {'1', '2', '3', ..., '8', '12', '34', '56', '78'}
            Select data from one spectral band or channel
            Default is '78' which combines band 7/8 to SWIR detector layout
        data       :  array-like
            data to be written with same dimensions as dataset "msm_dset"

        """
        fillvalue = float.fromhex('0x1.ep+122')

        if not self.__rw:
            raise PermissionError('read/write access required')

        if not self.__msm_path:
            return

        if not isinstance(band, str):
            raise TypeError('band must be a string')

        if band not in self.bands:
            raise ValueError('band not found in product')

        # skip row257 from the SWIR detector
        rows = None
        if int(band[0]) > 6:
            rows = [0, -1]

        # list potential names of the dataset dimensions
        time_list = ['time', 'scanline']
        row_list = ['width', 'pixel', 'ground_pixel']
        column_list = ['height', 'spectral_channel']

        indx = 0
        for ii in band:
            for dset_grp in ['OBSERVATIONS', 'ANALYSIS', '']:
                msm_path = str(self.__msm_path).replace('%', ii)
                ds_path = PurePosixPath(msm_path, dset_grp, msm_dset)
                if str(ds_path) not in self.fid:
                    continue
                dset = self.fid[str(ds_path)]

                data_sel = ()
                for xx in range(dset.ndim):
                    dim_name = PurePosixPath(dset.dims[xx][0].name).name

                    if len(dset.dims[xx][0][:]) == 1:
                        data_sel += (0,)
                    elif dim_name in time_list:
                        data_sel += (slice(None),)
                    elif dim_name in row_list:
                        if rows is None:
                            data_sel += (slice(None),)
                        else:
                            data_sel += (slice(*rows),)
                    elif dim_name in column_list:
                        if len(band) == 2:
                            jj = data.ndim-1
                            data = np.stack(np.split(data, 2, axis=jj))
                        data_sel += (slice(None),)
                    else:
                        raise ValueError

                if len(band) == 2:
                    if dset.attrs['_FillValue'] == fillvalue:
                        data[indx, np.isnan(data[indx, ...])] = fillvalue
                    dset[data_sel] = data[indx, ...]
                    indx += 1
                else:
                    if dset.attrs['_FillValue'] == fillvalue:
                        data[np.isnan(data)] = fillvalue
                    dset[data_sel] = data

                self.__patched_msm.append(ds_path)



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Implement function to obtain path to test-data

Copyright (c) 2016-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from os import environ
from pathlib import Path


def get_data_dir():
    """
    Obtain directory with test datasets

    Limited to UNIX/Linux/MacOS operating systems

    This module checks if the following directories are available:
     - /data/$USER/pys5p-data
     - /Users/$USER/pys5p-data
     - environment variable PYS5P_DATA_DIR

    It expects the data to be organized in the sub-directories:
     - CKD which should contain the SWIR dpqf CKD
     - OCM which should contain at least one directory of an on-ground
       calibration measurement with one or more OCAL LX products.
     - L1B which should contain at least one offline calibration, irradiance
       and radiance product.
     - ICM which contain at least one inflight calibration product.
    """
    try:
        user = environ['USER']
    except KeyError:
        print('*** Fatal: environment variable USER not set')
        return None

    guesses_data_dir = ['/data/{}/pys5p-data'.format(user),
                        '/Users/{}/pys5p-data'.format(user)]

    try:
        _ = environ['PYS5P_DATA_DIR']
    except KeyError:
        pass
    else:
        guesses_data_dir.append(environ['PYS5P_DATA_DIR'])

    for key in guesses_data_dir:
        if Path(key).is_dir():
            return key

    raise FileNotFoundError



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Performs unit-tests on class CKDio

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import Path
from pys5p.ckd_io import CKDio

def main():
    """
    """
    if Path('/nfs/Tropomi/share/ckd').is_dir():
        ckd_dir = '/nfs/Tropomi/share/ckd'
    elif Path('/data/richardh/Tropomi/share/ckd').is_dir():
        ckd_dir = '/data/richardh/Tropomi/share/ckd'
    else:
        ckd_dir = '/data2/richardh/Tropomi/share/ckd'

    print(ckd_dir)
    with CKDio(ckd_dir, ckd_version=1) as ckd:
        print(ckd.ckd_file)
        for meth in dir(ckd):
            if (meth.startswith('_')
                or meth.startswith('ckd')
                or meth in ('close', 'fid')):
                continue
            print('-------------------------', meth,
                  '-------------------------')
            print(meth, getattr(ckd, meth)())

    with CKDio(ckd_dir, ckd_version=2) as ckd:
        print(ckd.ckd_file)
        for meth in dir(ckd):
            if (meth.startswith('_')
                or meth.startswith('ckd')
                or meth in ('close', 'fid')):
                continue
            print('-------------------------', meth,
                  '-------------------------')
            print(meth, getattr(ckd, meth)())

# - main code --------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Performs unit-tests on S5Pplot methods: draw_signal, draw_quality,
   draw_cmp_images, draw_trend1d and draw_lines

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta
import numpy as np

from pys5p.lib.plotlib import FIGinfo
from pys5p.s5p_msm import S5Pmsm
from pys5p.s5p_plot import S5Pplot


def get_test_data(data_sel=None, xy_min=-5, xy_max=5, delta=0.01, error=0):
    """
    Generate synthetic data to simulate a square-detector image 
    """
    if data_sel is None:
        data_sel = [(), ()]

    res = np.arange(xy_min, xy_max, delta)
    xmesh, ymesh = np.meshgrid(res[data_sel[1]], res[data_sel[0]])
    zz1 = np.exp(-xmesh ** 2 - ymesh ** 2)
    zz2 = np.exp(-(xmesh - 1) ** 2 - (ymesh - 1) ** 2)
    data = (zz1 - zz2) * 2
    data += np.random.default_rng().normal(0., error, data.shape)

    msm = S5Pmsm(data)
    msm.error = np.full(data.shape, error)
    msm.set_units('Volt')
    msm.set_long_name('bogus data')

    return msm


# --------------------------------------------------
def run_draw_signal(plot):
    """
    Run unit tests on S5Pplot::draw_signal
    """
    print('Run unit tests on S5Pplot::draw_signal')

    msm = get_test_data(error=.1)
    msm_ref = get_test_data(error=0.025)

    nlines = 40 # 35
    fig_info_in = FIGinfo('right')
    for ii in range(nlines):
        fig_info_in.add('line {:02d}'.format(ii), 5 * '0123456789')

    plot.draw_signal(msm,
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='method=data; aspect=1; fig_pos=above')
    plot.draw_signal(msm, msm_ref.value, method='diff',
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='method=diff; aspect=1; fig_pos=above')
    plot.draw_signal(msm, msm_ref.value, method='ratio',
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='method=ratio; aspect=1; fig_pos=above')
    plot.draw_signal(msm, msm_ref, method='ratio_unc',
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='method=ratio_unc; aspect=1; fig_pos=above')
    plot.draw_signal(msm, method='error',
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='method=error; aspect=1; fig_pos=above')
    plot.draw_signal(msm, fig_info=fig_info_in.copy(),
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=1; fig_pos=right')

    msm = get_test_data(data_sel=np.s_[500-250:500+250, :], error=.1)
    plot.draw_signal(msm,
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=2; fig_pos=above')
    plot.draw_signal(msm, fig_info=fig_info_in.copy(),
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=2; fig_pos=right')

    msm = get_test_data(data_sel=np.s_[500-125:500+125, 500-375:500+375], error=.1)
    plot.draw_signal(msm,
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=3; fig_pos=above')
    plot.draw_signal(msm,
                     fig_info=fig_info_in.copy(),
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=3; fig_pos=right')

    msm = get_test_data(data_sel=np.s_[500-128:500+128, :], error=.1)
    plot.draw_signal(msm,
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=4; fig_pos=above')
    plot.draw_signal(msm, fig_info=fig_info_in.copy(),
                     title='Unit test of S5Pplot [draw_signal]',
                     sub_title='aspect=4; fig_pos=right')


def run_draw_cmp_swir(plot):
    """
    Run unit tests on S5Pplot::draw_cmp_swir
    """
    print('Run unit tests on S5Pplot::draw_cmp_swir')

    msm = get_test_data(data_sel=np.s_[500-128:500+128, :], error=.1)
    msm_ref = get_test_data(data_sel=np.s_[500-128:500+128, :], error=.025)

    plot.draw_cmp_swir(msm, msm_ref.value,
                       title='Unit test of S5Pplot [draw_cmp_images]',
                       sub_title='test image')
    plot.draw_cmp_swir(msm, msm_ref.value, add_residual=False,
                       title='Unit test of S5Pplot [draw_cmp_images]',
                       sub_title='test image')
    plot.draw_cmp_swir(msm, msm_ref.value, add_model=False,
                       title='Unit test of S5Pplot [draw_cmp_images]',
                       sub_title='test image')


def run_draw_trend1d(plot):
    """
    Run unit tests on S5Pplot::draw_trend1d
    """
    print('Run unit tests on S5Pplot::draw_trend1d')
    xx = np.arange(200) / 100
    hk_params = [
        ('detector_temp', 'SWIR detector temperature', 'K', np.float32),
        ('grating_temp', 'SWIR grating temperature', 'K', np.float32),
        ('obm_temp', 'SWIR OBM temperature', 'K', np.float32)]
    hk_dtype = np.dtype(
        [(parm[0], parm[3]) for parm in hk_params])
    hk_data = S5Pmsm(np.zeros(200, dtype=hk_dtype))
    hk_data.error = np.zeros((200, 2), dtype=hk_dtype)
    data = 140. + (100 - np.arange(200)) / 1000
    hk_data.value['detector_temp'][:] = data
    hk_data.error['detector_temp'][:, 0] = data - .0125
    hk_data.error['detector_temp'][:, 1] = data + .0075
    data = 202.1 + (100 - np.arange(200)) / 250
    hk_data.value['grating_temp'][:] = data
    hk_data.error['grating_temp'][:, 0] = data - .15
    hk_data.error['grating_temp'][:, 1] = data + .175
    data = 208.2 + (100 - np.arange(200)) / 750
    hk_data.value['obm_temp'][:] = data
    hk_data.error['obm_temp'][:, 0] = data - .15
    hk_data.error['obm_temp'][:, 1] = data + .175
    hk_data.set_units([parm[2] for parm in hk_params])
    hk_data.set_long_name([parm[1] for parm in hk_params])

    plot.draw_trend1d(S5Pmsm(np.sin(xx * np.pi)),
                      title='Unit test of S5Pplot [draw_trend1d]',
                      sub_title='one dataset, no house-keeping')
    plot.draw_trend1d(S5Pmsm(np.sin(xx * np.pi)),
                      msm2=S5Pmsm(np.cos(xx * np.pi)),
                      title='Unit test of S5Pplot [draw_trend1d]',
                      sub_title='two datasets, no house-keeping')
    hk_keys = [parm[0] for parm in hk_params]
    plot.draw_trend1d(S5Pmsm(np.sin(xx * np.pi)),
                      hk_data=hk_data, hk_keys=hk_keys[0:2],
                      msm2=S5Pmsm(np.cos(xx * np.pi)),
                      title='Unit test of S5Pplot [draw_trend1d]',
                      sub_title='two datasets and house-keeping')
    plot.draw_trend1d(S5Pmsm(np.sin(xx * np.pi)),
                      hk_data=hk_data, hk_keys=hk_keys,
                      msm2=S5Pmsm(np.cos(xx * np.pi)),
                      title='Unit test of S5Pplot [draw_trend1d]',
                      sub_title='two datasets and house-keeping')


def run_draw_lines(plot):
    """
    Run unit tests on S5Pplot::draw_lines
    """
    print('Run unit tests on S5Pplot::draw_lines')
    xx = np.arange(200) / 100
    plot.draw_lines(xx, np.sin(xx * np.pi), color=0,
                    label='sinus', marker='o', linestyle='-')
    plot.draw_lines(xx, np.cos(xx * np.pi), color=1,
                    label='cosinus', marker='o', linestyle='-')
    plot.draw_lines(None, None, ylim=[-1.05, 1.05],
                    xlabel='x-axis [Pi]', ylabel='y-axis',
                    title='Unit test of S5Pplot [draw_lines]',
                    sub_title='draw_lines [no time_axis]')

    xx = np.arange(500) / 100
    plot.draw_lines(xx, np.sin(xx * np.pi), color=0,
                    label='sinus', marker='o', linestyle='-')
    plot.draw_lines(xx, np.cos(xx * np.pi), color=1,
                    label='cosinus', marker='o', linestyle='-')
    plot.draw_lines(None, None, ylim=[-1.05, 1.05],
                    xlabel='x-axis [Pi]', ylabel='y-axis',
                    title='Unit test of S5Pplot [draw_lines]',
                    sub_title='draw_lines [no time_axis]')

    customdate = datetime(2016, 1, 1, 13, 0, 0)
    yy = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20]
    xx = [customdate + timedelta(hours=i, minutes=4*i)
          for i in range(len(yy))]
    plot.draw_lines(xx, yy, color=0, label='mydata',
                    marker='o', linestyle='-')
    plot.draw_lines(None, None, title='Unit test of S5Pplot [draw_lines]',
                    xlabel='x-axis', ylabel='y-axis',
                    sub_title='draw_lines [time_axis]')

def run_draw_quality(plot):
    """
    Run unit tests on S5Pplot::draw_quality
    """
    print('Run unit tests on S5Pplot::draw_quality')
    row = np.linspace(0, 1., 1000)
    ref_data = np.repeat(row[None, :], 256, axis=0)

    data = ref_data.copy()
    data[125:175, 30:50] = 0.4
    data[125:175, 50:70] = 0.95
    data[75:125, 250:300] = 0.05
    data[125:175, 600:650] = 0.95
    data[75:125, 900:925] = 0.05
    data[75:125, 825:875] = 0.4

    plot.draw_quality(data,
                      title='Unit test of S5Pplot [draw_quality]',
                      sub_title='no reference')

    plot.draw_quality(data, ref_data=ref_data,
                      title='Unit test of S5Pplot [draw_quality]',
                      sub_title='with reference')


def run_draw_qhist(plot):
    """
    Run unit tests on S5Pplot::draw_qhist
    """
    print('Run unit tests on S5Pplot::draw_qhist')
    buff0 = np.repeat(0.9 + np.random.rand(1000) / 10, 56)
    buff1 = np.repeat(np.random.rand(1000) / 10, 10)
    buff2 = 0.1 + np.random.rand(1000) / 10
    buff3 = np.repeat(0.2 + np.random.rand(1000) / 10, 2)
    buff4 = np.repeat(0.3 + np.random.rand(1000) / 10, 3)
    buff5 = np.repeat(0.4 + np.random.rand(1000) / 10, 4)
    buff6 = np.repeat(0.5 + np.random.rand(1000) / 10, 8)
    buff7 = np.repeat(0.6 + np.random.rand(1000) / 10, 12)
    buff8 = np.repeat(0.7 + np.random.rand(1000) / 10, 20)
    buff9 = np.repeat(0.8 + np.random.rand(1000) / 10, 40)
    buffa = np.repeat(0.9 + np.random.rand(1000) / 10, 100)
    data = S5Pmsm(np.concatenate((buff0, buff1, buff2, buff3, buff4,
                                  buff5, buff6, buff7, buff8, buff9,
                                  buffa)).reshape(256, 1000))
    data_dict = {}
    data.set_long_name('pixel-quality map')
    data_dict['dpqm'] = data.copy()
    data.set_long_name('pixel-quality map (dark)', force=True)
    data_dict['dpqm_dark'] = data.copy()
    data.set_long_name('pixel-quality map (noise average)', force=True)
    data_dict['dpqm_noise'] = data.copy()
    data.set_long_name('pixel-quality map (noise variance)', force=True)
    data_dict['dpqm_noise_var'] = data
    plot.draw_qhist(data_dict,
                    title='Unit test of S5Pplot [draw_qhist]')


# --------------------------------------------------
def main():
    """
    main function
    """
    plot = S5Pplot('unit_test_s5p_plot.pdf')
    check_draw_signal = True
    check_draw_cmp_images = True
    check_draw_quality = False
    check_draw_qhist = False
    check_draw_trend1d = False
    check_draw_lines = False

    # ---------- UNIT TEST: draw_signal ----------
    if check_draw_signal:
        run_draw_signal(plot)

    # ---------- UNIT TEST: draw_cmp_images ----------
    if check_draw_cmp_images:
        run_draw_cmp_swir(plot)

    # ---------- UNIT TEST: draw_quality ----------
    if check_draw_quality:
        run_draw_quality(plot)

    # ---------- UNIT TEST: draw_qhist ----------
    if check_draw_qhist:
        run_draw_qhist(plot)

    # ---------- UNIT TEST: draw_lines ----------
    if check_draw_trend1d:
        run_draw_trend1d(plot)

    # ---------- UNIT TEST: draw_lines ----------
    if check_draw_lines:
        run_draw_lines(plot)

    # close figure
    plot.close()


# - main code --------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pys5p

https://github.com/rmvanhees/pys5p.git

Implement some function for error propagation

Copyright (c) 2017-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np


def unc_sum(sigma_a, sigma_b):
    """
    absolute error for a +/- b is sqrt(sigma_a^2 + sigma_b^2)
    """
    if sigma_a.shape != sigma_b.shape:
        raise TypeError('dimensions of sigma_a and sigma are not the same')

    mask = np.isfinite(sigma_a) & np.isfinite(sigma_b)

    if np.sum(mask) == sigma_a.size:
        return np.sqrt(sigma_a ** 2 + sigma_b ** 2)

    res = np.full(sigma_a.shape, np.nan)
    res[mask] = np.sqrt(sigma_a[mask] ** 2 + sigma_b[mask] ** 2)

    return res


def unc_div(value_a, sigma_a, value_b, sigma_b):
    """
    absolute error for a / b is sqrt((sigma_a/a)^2 + (sigma_b/b)^2) * a/b
    """
    if not value_a.shape == value_b.shape == sigma_a.shape == sigma_b.shape:
        raise TypeError('dimensions of input arrays are not the same')

    mask = (np.isfinite(value_a) & np.isfinite(sigma_a)
            & np.isfinite(value_b) & np.isfinite(sigma_b))

    if np.sum(mask) == sigma_a.size:
        return (value_a / value_b) * np.sqrt((sigma_a / value_a) ** 2
                                             + (sigma_b / value_b) ** 2)

    res = np.full(sigma_a.shape, np.nan)
    res[mask] = ((value_a[mask] / value_b[mask])
                 * np.sqrt((sigma_a[mask] / value_a[mask]) ** 2
                           + (sigma_b[mask] / value_b[mask]) ** 2))
    return res



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Provides access to the S5P Tropomi CKD (static and dynamic)

ToDo
----
 - access to UVN CKD, still incomplete
 - identify latest Static CKD product, e.g. using the validity period

Copyright (c) 2018-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import Path, PosixPath

import h5py
import numpy as np

from .s5p_msm import S5Pmsm

# - global parameters ------------------------------


# - class definition -------------------------------
class CKDio():
    """
    Read Tropomi CKD from the Static CKD product or from dynamic CKD products

    Attributes
    ----------
    ckd_dir : pathlib.Path
    ckd_version : int
    ckd_file : pathlib.Path
    ckd_dyn_file : pathlib.Path
    fid : h5py.File

    Methods
    -------
    close()
       Close resources.
    creation_time()
       Returns datetime when the L1b product was created.
    creator_version()
       Returns version of Tropomi L01B processor used to generate this procuct.
    get_param(ds_name=None, band='7')
       Returns value(s) of a CKD parameter from the Static CKD product.
    memory(bands='78')
       Returns memory CKD, SWIR only.
    dn2v(bands='78')
       Returns digital number to Volt CKD, SWIR only.
    voltage_to_charge(bands='78')
       Returns Voltage to Charge CKD, SWIR only.
    prnu(bands='78')
       Returns Pixel Response Non-Uniformity (PRNU).
    absirr(qvd=1, bands='78')
       Returns absolute irradiance responsivity.
    relirr(qvd=1, bands='78')
       Returns relative irradiance correction.
    absrad(bands='78')
       Returns absolute radiance responsivity.
    wavelength(bands='78')
       Returns wavelength CKD.
    saa()
       Returns definition of the SAA region.
    offset(bands='78')
       Returns offset CKD, SWIR only.
    darkflux(bands='78')
       Returns dark-flux CKD, SWIR only.
    dpqf(threshold=None, bands='78')
       Returns Detector Pixel Quality Mask (boolean), SWIR only.
    pixel_quality(bands='78')
       Returns Detector Pixel Quality Mask (float [0, 1]), SWIR only.
    noise(bands='78')
       Returns pixel noise CKD, SWIR only
    saturation(bands='78')
       Returns saturation values (pre-offset), SWIR only

    Notes
    -----
    Not all CKD are defined or derived for all bands.
    You can request a CKD for one band or for a channel (bands: '12', '34',
    '56', '78'). Do not mix bands from different channels

    The option to have dynamic CKD is not used for the Tropomi mission, only
    for S/W version 1 a dynamic CKD product is defined. This product contained
    the OCAL CKD and was not updated automatically. For version 2, all CKD are
    stored in one product, where some CKD have a time-axis to correct any
    in-flight degradation.

    Therefore, the logic to find a CKD is implemented as follows:
    1) ckd_dir, defines the base directory to search for the CKD products
    (see below)
    2) ckd_file, defines the full path to (static) CKD product;
    (version 1) any product with dynamic CKD has to be in the same directory

    Version 1:
    * Static CKD are stored in one file: glob('*_AUX_L1_CKD_*')
    * Dynamic CKD are stored in two files:
      - UVN, use glob('*_ICM_CKDUVN_*')
      - SWIR, use glob('*_ICM_CKDSIR_*')

    Version 2+:
    * All CKD in one file: glob('*_AUX_L1_CKD_*')
    * Dynamic CKD are empty

    Examples
    --------
    """
    def __init__(self, ckd_dir=None, ckd_version=1, ckd_file=None):
        """
        Initialize access to a Tropomi Static CKD product

        Parameters
        ----------
        ckd_dir :  str, optional
           Directory where the CKD files are stored,
           default='/nfs/Tropomi/share/ckd'
        ckd_version :  int, optional
           Version of the CKD, default=1
        ckd_file : str, optional
           Name of the CKD file, default=None then the CKD file is searched
           in the directory ckd_dir with ckd_version in the glob-string
        """
        if ckd_dir is None:
            ckd_dir = '/nfs/Tropomi/share/ckd'
        self.ckd_version = max(1, ckd_version)
        self.ckd_dyn_file = None

        # define path to CKD product
        if ckd_file is None:
            if not Path(ckd_dir).is_dir():
                raise FileNotFoundError(
                    'Not found CKD directory: {}'.format(ckd_dir))
            self.ckd_dir = Path(ckd_dir)
            glob_str = '*_AUX_L1_CKD_*_*_00000_{:02d}_*_*.h5'.format(
                self.ckd_version)
            if (self.ckd_dir / 'static').is_dir():
                res = sorted((self.ckd_dir / 'static').glob(glob_str))
            else:
                res = sorted(self.ckd_dir.glob(glob_str))
            if not res:
                raise FileNotFoundError('Static CKD product not found')
            self.ckd_file = res[-1]
        else:
            if not Path(ckd_file).is_file():
                raise FileNotFoundError(
                    'Not found CKD file: {}'.format(ckd_file))
            self.ckd_dir = Path(ckd_file).parent
            self.ckd_file = Path(ckd_file)

        # obtain path to dynamic CKD product (version 1, only)
        if self.ckd_version == 1:
            if (self.ckd_dir / 'dynamic').is_dir():
                res = sorted((self.ckd_dir / 'dynamic').glob('*_ICM_CKDSIR_*'))
            else:
                res = sorted(self.ckd_dir.glob('*_ICM_CKDSIR_*'))
            if res:
                self.ckd_dyn_file = res[-1]

        # open access to CKD product
        self.fid = h5py.File(self.ckd_file, "r")

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self) -> None:
        """
        Make sure that we close all resources
        """
        if self.fid is not None:
            self.fid.close()

    def creation_time(self) -> str:
        """
        Returns datetime when the L1b product was created
        """
        if self.ckd_version == 2:
            attr = self.fid['METADATA'].attrs['production_datetime']
        else:
            group = PosixPath('METADATA', 'earth_explorer_header',
                              'fixed_header', 'source')
            attr = self.fid[str(group)].attrs['Creator_Date'][0]

        if isinstance(attr, bytes):
            attr = attr.decode('ascii')
        return attr

    def creator_version(self) -> str:
        """
        Returns version of Tropomi L01B processor used to generate this procuct
        """
        group = PosixPath('METADATA', 'earth_explorer_header', 'fixed_header')
        attr = self.fid[str(group)].attrs['File_Version']
        if self.ckd_version == 1:
            attr = attr[0]
        if isinstance(attr, bytes):
            attr = attr.decode('ascii')
        return attr

    # ---------- static CKD's ----------
    def get_param(self, ds_name=None, band='7'):
        """
        Returns value(s) of a CKD parameter from the Static CKD product.

        Parameters
        ----------
        ds_name  :  string
           Name of the HDF5 dataset, default='pixel_full_well'
        band     : string
           Band identifier '1', '2', ..., '8', default='7'

        Returns
        -------
        numpy.ndarray or scalar

        Notes
        -----
        Datasets of size=1 are return as scalar

        Handy function for scalar HDF5 datasets, such as:
         - dc_reference_temp
         - dpqf_threshold
         - pixel_full_well
         - pixel_fw_flag_thresh
        """
        if ds_name is None:
            ds_name = 'pixel_full_well'

        if not 1 <= int(band) <= 8:
            raise ValueError('band must be between and 1 and 8')

        if ds_name not in self.fid['/BAND{}'.format(band)]:
            raise ValueError('dataset not available')

        full_name = '/BAND{}/{}'.format(band, ds_name)
        # pylint: disable=no-member
        if self.fid[full_name].size == 1:
            return self.fid[full_name][0]

        return self.fid[full_name][:]

    def memory(self, bands='78') -> dict:
        """
        Returns memory CKD, SWIR only

        Note SWIR row 257 is excluded
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('Memory CKD only available for SWIR')

        long_name = 'SWIR memory CKD ({})'
        ckd_parms = ['mem_lin_neg_swir', 'mem_lin_pos_swir',
                     'mem_qua_neg_swir', 'mem_qua_pos_swir']
        ckd = {}
        for key in ckd_parms:
            for band in bands:
                if key not in ckd:
                    ckd[key] = S5Pmsm(self.fid['/BAND{}/{}'.format(band, key)],
                                      datapoint=True, data_sel=np.s_[:-1, :])
                    ckd[key].set_long_name(long_name.format(key))
                else:
                    buff = S5Pmsm(self.fid['/BAND{}/{}'.format(band, key)],
                                  datapoint=True, data_sel=np.s_[:-1, :])
                    ckd[key].concatenate(buff, axis=1)

            ckd[key].set_fillvalue()
            ckd[key].fill_as_nan()

        return ckd

    def dn2v(self, bands='78'):
        """
        Returns digital number to Volt CKD, SWIR only

        Notes
        -----
        The DN2V factor has no error attached to it.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('DN2V factor is only available for SWIR')

        ckd = S5Pmsm(self.fid['/BAND7/dn2v_factor_swir'])
        ckd8 = S5Pmsm(self.fid['/BAND8/dn2v_factor_swir'])
        ckd.value[2:] = ckd8.value[2:]
        ckd.set_long_name('SWIR DN2V factor')
        return ckd

    # def non_linearity(self, bands='12'):
    #    """
    #    Returns non-linearity CKD, UVN only
    #    """
    #    if len(bands) > 2:
    #        raise ValueError('read per band or channel, only')
    #    if '7' in bands or '8' in bands:
    #        raise ValueError('non-linearity only available for UVN')
    #    else:
    #        raise NotImplementedError('not implemented, yet')

    def voltage_to_charge(self, bands='78'):
        """
        Returns Voltage to Charge CKD, SWIR only

        Notes
        -----
        The V2C CKD has no error attached to it.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('Voltage to Charge only available for SWIR')

        ckd = S5Pmsm(self.fid['/BAND7/v2c_factor_swir'], datapoint=True)
        ckd.set_long_name('SWIR voltage to charge CKD')
        return ckd

    def prnu(self, bands='78'):
        """
        Returns Pixel Response Non-Uniformity (PRNU)

        Notes
        -----
        * SWIR row 257 is excluded.
        * The PRNU-CKD has no error attached to it (always zero).
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')

        data_sel = None
        if '7' in bands or '8' in bands:
            data_sel = np.s_[:-1, :]
            long_name = 'SWIR PRNU CKD'
        elif '5' in bands or '6' in bands:
            long_name = 'NIR PRNU CKD'
        elif '3' in bands or '4' in bands:
            long_name = 'VIS PRNU CKD'
        else:
            long_name = 'UV PRNU CKD'

        ckd = None
        for band in bands:
            if not ckd:
                ckd = S5Pmsm(self.fid['/BAND{}/PRNU'.format(band)],
                             datapoint=True, data_sel=data_sel)
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid['/BAND{}/PRNU'.format(band)],
                              datapoint=True, data_sel=data_sel)
                ckd.concatenate(buff, axis=1)

            ckd.set_fillvalue()
        return ckd

    def absirr(self, qvd=1, bands='78'):
        """
        Returns absolute irradiance responsivity

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')

        data_sel = None
        if '7' in bands or '8' in bands:
            data_sel = np.s_[:-1, :]
            long_name = 'SWIR absolute irradiance CKD (QVD={})'.format(qvd)
        elif '5' in bands or '6' in bands:
            long_name = 'NIR absolute irradiance CKD (QVD={})'.format(qvd)
        elif '3' in bands or '4' in bands:
            long_name = 'VIS absolute irradiance CKD (QVD={})'.format(qvd)
        else:
            long_name = 'UV absolute irradiance CKD (QVD={})'.format(qvd)

        ckd = None
        for band in bands:
            dsname = '/BAND{}/abs_irr_conv_factor_qvd{}'.format(band, qvd)
            if not ckd:
                ckd = S5Pmsm(self.fid[dsname],
                             datapoint=True, data_sel=data_sel)
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname],
                              datapoint=True, data_sel=data_sel)
                ckd.concatenate(buff, axis=1)

            ckd.set_fillvalue()

        return ckd

    def relirr(self, qvd=1, bands='78'):
        """
        Returns relative irradiance correction

        Parameters
        ----------
         - qvd   : integer
                   select data from QVD1 or QVD2.
                   Default QVD1
         - bands : string
                   select CKD for one band or a channel. Channels can be
                   selected by bands equals '12', '34', '56' or '78'.
                   Default SWIR channel

        Returns
        -------
        dictionary with keys:
         - band           :    Tropomi spectral band ID
         - mapping_cols   :    coarse irregular mapping of the columns
         - mapping_rows   :    coarse irregular mapping of the rows
         - cheb_coefs     :    chebyshev parameters for elevation and azimuth
                               for pixels on a coarse irregular grid

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')

        res = ()
        for band in bands:
            ckd = {}
            ckd['band'] = int(band)

            dsname = '/BAND{}/rel_irr_coarse_mapping_vert'.format(band)
            ckd['mapping_rows'] = self.fid[dsname][:].astype(int)

            dsname = '/BAND{}/rel_irr_coarse_mapping_hor'.format(band)
            # pylint: disable=no-member
            mapping_hor = self.fid[dsname][:].astype(int)
            mapping_hor[mapping_hor > 1000] -= 2**16
            ckd['mapping_cols'] = mapping_hor

            dsname = '/BAND{}/rel_irr_coarse_func_cheb_qvd{}'.format(band, qvd)
            ckd['cheb_coefs'] = self.fid[dsname]['coefs'][:]
            res += (ckd,)

        return res

    def absrad(self, bands='78'):
        """
        Returns absolute radiance responsivity

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')

        data_sel = None
        if '7' in bands or '8' in bands:
            data_sel = np.s_[:-1, :]
            long_name = 'SWIR absolute radiance CKD'
        elif '5' in bands or '6' in bands:
            long_name = 'NIR absolute radiance CKD'
        elif '3' in bands or '4' in bands:
            long_name = 'VIS absolute radiance CKD'
        else:
            long_name = 'UV absolute radiance CKD'

        ckd = None
        for band in bands:
            dsname = '/BAND{}/abs_rad_conv_factor'.format(band)
            if not ckd:
                ckd = S5Pmsm(self.fid[dsname],
                             datapoint=True, data_sel=data_sel)
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname],
                              datapoint=True, data_sel=data_sel)
                ckd.concatenate(buff, axis=1)

            ckd.set_fillvalue()

        return ckd

    def wavelength(self, bands='78'):
        """
        Returns wavelength CKD

        Notes
        -----
        * SWIR row 257 is excluded.
        * The wavelength CKD has no error attached to it.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')

        data_sel = None
        if '7' in bands or '8' in bands:
            data_sel = np.s_[:-1, :]
            long_name = 'SWIR wavelength CKD'
        elif '5' in bands or '6' in bands:
            long_name = 'NIR wavelength CKD'
        elif '3' in bands or '4' in bands:
            long_name = 'VIS wavelength CKD'
        else:
            long_name = 'UV wavelength CKD'

        ckd = None
        for band in bands:
            dsname = '/BAND{}/wavelength_map'.format(band)
            if not ckd:
                ckd = S5Pmsm(self.fid[dsname],
                             datapoint=True, data_sel=data_sel)
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname],
                              datapoint=True, data_sel=data_sel)
                ckd.concatenate(buff, axis=1)

            ckd.set_fillvalue()

        return ckd

    def saa(self) -> dict:
        """
        Returns definition of the SAA region
        """
        saa_region = {}
        saa_region['lat'] = self.fid['saa_latitude'][:]
        saa_region['lon'] = self.fid['saa_longitude'][:]

        return saa_region

    # ---------- external CKD's ----------
    def offset(self, bands='78'):
        """
        Returns offset CKD, SWIR only

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('Offset CKD is only available for SWIR')

        ckd = None
        long_name = 'SWIR offset CKD'

        # try the Static CKD product, first
        for band in bands:
            dsname = '/BAND{}/analog_offset_swir'.format(band)
            if dsname not in self.fid:
                continue

            if not ckd:
                ckd = S5Pmsm(self.fid[dsname],
                             datapoint=True, data_sel=np.s_[:-1, :])
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname],
                              datapoint=True, data_sel=np.s_[:-1, :])
                ckd.concatenate(buff, axis=1)

        if ckd is not None:
            ckd.set_fillvalue()
            return ckd

        # try the dynamic CKD products
        with h5py.File(self.ckd_dyn_file, 'r') as fid:
            for band in bands:
                dsname = '/BAND{}/analog_offset_swir'.format(band)

                if not ckd:
                    ckd = S5Pmsm(fid[dsname],
                                 datapoint=True, data_sel=np.s_[:-1, :])
                    ckd.set_long_name(long_name)
                else:
                    buff = S5Pmsm(fid[dsname],
                                  datapoint=True, data_sel=np.s_[:-1, :])
                    ckd.concatenate(buff, axis=1)
        ckd.set_fillvalue()
        return ckd

    def darkflux(self, bands='78'):
        """
        Returns dark-flux CKD, SWIR only

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('Dark-flux CKD is only available for SWIR')

        ckd = None
        long_name = 'SWIR dark-flux CKD'

        # try the Static CKD product, first
        for band in bands:
            dsname = '/BAND{}/long_term_swir'.format(band)
            if dsname not in self.fid:
                continue

            if not ckd:
                ckd = S5Pmsm(self.fid[dsname],
                             datapoint=True, data_sel=np.s_[:-1, :])
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname],
                              datapoint=True, data_sel=np.s_[:-1, :])
                ckd.concatenate(buff, axis=1)

        if ckd is not None:
            ckd.set_fillvalue()
            return ckd

        # try the dynamic CKD products
        with h5py.File(self.ckd_dyn_file, 'r') as fid:
            for band in bands:
                dsname = '/BAND{}/long_term_swir'.format(band)

                if not ckd:
                    ckd = S5Pmsm(fid[dsname],
                                 datapoint=True, data_sel=np.s_[:-1, :])
                    ckd.set_long_name(long_name)
                else:
                    buff = S5Pmsm(fid[dsname],
                                  datapoint=True, data_sel=np.s_[:-1, :])
                    ckd.concatenate(buff, axis=1)

        ckd.set_fillvalue()
        return ckd

    def dpqf(self, threshold=None, bands='78'):
        """
        Returns Detector Pixel Quality Mask (boolean), SWIR only

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('pixel quality CKD is only available for SWIR')

        if threshold is None:
            threshold = self.fid['/BAND7/dpqf_threshold'][:]

        # try the Static CKD product, first
        if '/BAND7/dpqf_map' in self.fid:
            dset = self.fid['/BAND7/dpqf_map']
            dpqf_b7 = dset[:-1, :]
            dset = self.fid['/BAND8/dpqf_map']
            dpqf_b8 = dset[:-1, :]

            return np.hstack((dpqf_b7, dpqf_b8)) < threshold

        # try the dynamic CKD products
        with h5py.File(self.ckd_dyn_file, 'r') as fid:
            if threshold is None:
                threshold = fid['/BAND7/dpqf_threshold'][:]

            dset = fid['/BAND7/dpqf_map']
            dpqf_b7 = dset[:-1, :]
            dset = fid['/BAND8/dpqf_map']
            dpqf_b8 = dset[:-1, :]

        return np.hstack((dpqf_b7, dpqf_b8)) < threshold

    def pixel_quality(self, bands='78'):
        """
        Returns Detector Pixel Quality Mask (float [0, 1]), SWIR only

        Notes
        -----
        SWIR row 257 is excluded.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('pixel quality CKD is only available for SWIR')

        ckd = None
        long_name = 'SWIR pixel-quality CKD'

        # try the Static CKD product, first
        for band in bands:
            dsname = '/BAND{}/dpqf_map'.format(band)
            if dsname not in self.fid:
                continue

            if not ckd:
                ckd = S5Pmsm(self.fid[dsname], data_sel=np.s_[:-1, :])
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname], data_sel=np.s_[:-1, :])
                ckd.concatenate(buff, axis=1)

        if ckd is not None:
            return ckd

        # try the dynamic CKD products
        with h5py.File(self.ckd_dyn_file, 'r') as fid:
            for band in bands:
                dsname = '/BAND{}/dpqf_map'.format(band)

                if not ckd:
                    ckd = S5Pmsm(fid[dsname], data_sel=np.s_[:-1, :])
                    ckd.set_long_name(long_name)
                else:
                    buff = S5Pmsm(fid[dsname], data_sel=np.s_[:-1, :])
                    ckd.concatenate(buff, axis=1)

        return ckd

    def noise(self, bands='78'):
        """
        Returns noise CKD, SWIR only

        Notes
        -----
        * SWIR row 257 is excluded.
        * The noise CKD has no error attached to it.
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('noise CKD is only available for SWIR')

        ckd = None
        long_name = 'SWIR noise CKD'

        # try the Static CKD product, first
        for band in bands:
            dsname = '/BAND{}/readout_noise_swir'.format(band)
            if dsname not in self.fid:
                continue

            if not ckd:
                ckd = S5Pmsm(self.fid[dsname], data_sel=np.s_[:-1, :])
                ckd.set_long_name(long_name)
            else:
                buff = S5Pmsm(self.fid[dsname], data_sel=np.s_[:-1, :])
                ckd.concatenate(buff, axis=1)

        if ckd is not None:
            ckd.set_fillvalue()
            return ckd

        # try the dynamic CKD products
        ckd_file = self.ckd_dir / 'OCAL' / 'ckd.readnoise.detector4.nc'
        with h5py.File(ckd_file, 'r') as fid:
            for band in bands:
                dsname = '/BAND{}/readout_noise_swir'.format(band)

                if not ckd:
                    ckd = S5Pmsm(fid[dsname],
                                 datapoint=True, data_sel=np.s_[:-1, :])
                    ckd.set_long_name(long_name)
                else:
                    buff = S5Pmsm(fid[dsname],
                                  datapoint=True, data_sel=np.s_[:-1, :])
                    ckd.concatenate(buff, axis=1)

        ckd.set_fillvalue()
        return ckd

    def saturation(self, bands='78'):
        """
        Returns saturation values (pre-offset), SWIR only

        Notes
        -----
        * SWIR row 257 is excluded
        * The saturation CKD has no error attached to it (always zero)
        """
        if len(bands) > 2:
            raise ValueError('read per band or channel, only')
        if '7' not in bands and '8' not in bands:
            raise ValueError('saturation CKD is only available for SWIR')

        ckd = None
        long_name = 'SWIR saturation(pre-offset) CKD'
        ckd_file = (self.ckd_dir / 'OCAL'
                    / 'ckd.saturation_preoffset.detector4.nc')
        with h5py.File(ckd_file, 'r') as fid:
            for band in bands:
                dsname = '/BAND{}/saturation_preoffset'.format(band)

                if not ckd:
                    ckd = S5Pmsm(fid[dsname], data_sel=np.s_[:-1, :])
                    ckd.set_long_name(long_name)
                else:
                    buff = S5Pmsm(fid[dsname], data_sel=np.s_[:-1, :])
                    ckd.concatenate(buff, axis=1)

        ckd.set_fillvalue()
        return ckd



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

The class LV2io provides read access to S5p Tropomi S5P_OFFL_L2 products

Copyright (c) 2018-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta
from pathlib import Path

import h5py
from netCDF4 import Dataset
import numpy as np

from .s5p_msm import S5Pmsm

# - global parameters ------------------------------


# - local functions --------------------------------


# - class definition -------------------------------
class LV2io():
    """
    This class should offer all the necessary functionality to read Tropomi
    S5P_OFFL_L2 products

    Attributes
    ----------
    fid : h5py.File
    filename : string
    science_product : bool
    ground_pixel : int
    scanline : int

    Methods
    -------
    close()
       Close resources.
    get_attr(attr_name, ds_name=None)
       Obtain value of an HDF5 file attribute or dataset attribute.
    get_orbit()
       Returns reference orbit number
    get_algorithm_version()
       Returns version of the level-2 algorithm.
    get_processor_version()
       Returns version of the L12 processor used to generate this product.
    get_product_version()
       Returns version of the level-2 product
    get_coverage_time()
       Returns start and end of the measurement coverage time.
    get_creation_time()
       Returns creation date/time of the level-2 product.
    get_ref_time()
       Returns reference start time of measurements.
    get_delta_time()
       Returns offset from the reference start time of measurement.
    get_geo_data(geo_dsets=None)
       Returns data of selected datasets from the GEOLOCATIONS group.
    get_geo_bounds(extent=None, data_sel=None)
       Returns bounds of latitude/longitude as a mesh for plotting.
    get_dataset(name, data_sel=None, fill_as_nan=True)
       Read level-2 dataset from PRODUCT group.
    get_data_as_s5pmsm(name, data_sel=None, fill_as_nan=True, mol_m2=False)
       Read dataset from group PRODUCT/target_product group.

    Notes
    -----
    The Python h5py module can read the operational netCDF4 products without
    any problems, however, the SRON science products contain incompatible
    attributes. Thus should be fixed when more up-to-date netCDF software is
    used to generate the products. Currently, the Python netCDF4 module is
    used to read the science products.

    Examples
    --------
    """
    def __init__(self, lv2_product):
        """
        Initialize access to an S5P_L2 product

        Parameters
        ----------
        lv2_product :  string
           full path to S5P Tropomi level 2 product
        """
        if not Path(lv2_product).is_file():
            raise FileNotFoundError('{} does not exist'.format(lv2_product))

        # initialize class-attributes
        self.filename = lv2_product
        self.science_product = False

        # open LV2 product as HDF5 file
        self.fid = h5py.File(lv2_product, "r")
        try:
            science_inst = ['SRON Netherlands Institute for Space Research']

            if self.get_attr('institution') in science_inst:
                self.science_product = True
        except OSError:
            self.science_product = True

        if self.science_product:
            self.fid.close()
            self.fid = Dataset(lv2_product, "r", format="NETCDF4")

            self.ground_pixel = self.fid['/instrument/ground_pixel'][:].max()
            self.ground_pixel += 1
            self.scanline = self.fid['/instrument/scanline'][:].max()
            self.scanline += 1
            # alternative set flag sparse
            if self.fid['/instrument/scanline'].size % self.ground_pixel != 0:
                raise ValueError('not all scanlines are complete')
        else:
            self.ground_pixel = self.fid['/PRODUCT/ground_pixel'].size
            self.scanline = self.fid['/PRODUCT/scanline'].size

    def __repr__(self):
        class_name = type(self).__name__
        return '{}({!r})'.format(class_name, self.filename)

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self):
        """
        Close the product.
        """
        if self.fid is not None:
            self.fid.close()

    # -------------------------
    def __h5_attr(self, attr_name, ds_name):
        """
        read attributes from operational products using hdf5
        """
        if ds_name is not None:
            dset = self.fid['/PRODUCT/{}'.format(ds_name)]
            if attr_name not in dset.attrs.keys():
                return None

            attr = dset.attrs[attr_name]
        else:
            if attr_name not in self.fid.attrs:
                return None

            attr = self.fid.attrs[attr_name]

        if isinstance(attr, bytes):
            return attr.decode('ascii')

        return attr

    def __nc_attr(self, attr_name, ds_name):
        """
        read attributes from science products using netCDF4
        """
        if ds_name is not None:
            for grp_name in ['target_product', 'side_product', 'instrument']:
                if grp_name not in self.fid.groups:
                    continue

                if ds_name not in self.fid[grp_name].variables:
                    continue

                dset = self.fid['/{}/{}'.format(grp_name, ds_name)]
                if attr_name in dset.ncattrs():
                    return dset.getncattr(attr_name)

            return None

        if attr_name not in self.fid.ncattrs():
            return None

        return self.fid.getncattr(attr_name)

    def get_attr(self, attr_name, ds_name=None):
        """
        Obtain value of an HDF5 file attribute or dataset attribute

        Parameters
        ----------
        attr_name : string
           name of the attribute
        ds_name   : string (optional)
           name of dataset, default is to read the product attributes
        """
        if self.science_product:
            return self.__nc_attr(attr_name, ds_name)

        return self.__h5_attr(attr_name, ds_name)

    # -------------------------
    def get_orbit(self) -> int:
        """
        Returns reference orbit number
        """
        if self.science_product:
            return int(self.__nc_attr('orbit', 'l1b_file'))

        return self.__h5_attr('orbit', None)[0]

    # -------------------------
    def get_algorithm_version(self) -> str:
        """
        Returns version of the level 2 algorithm
        """
        return self.get_attr('algorithm_version')

    # -------------------------
    def get_processor_version(self) -> str:
        """
        Returns version of the L12 processor
        """
        return self.get_attr('processor_version')

    # -------------------------
    def get_product_version(self):
        """
        Returns version of the level 2 product
        """
        return self.get_attr('product_version')

    # -------------------------
    def get_coverage_time(self):
        """
        Returns start and end of the measurement coverage time
        """
        res1 = self.get_attr('time_coverage_start')
        res2 = self.get_attr('time_coverage_end')
        return (res1, res2)

    # -------------------------
    def get_creation_time(self):
        """
        Returns creation date/time of the level 2 product
        """
        return self.get_attr('date_created')

    # ---------- Functions using data in the PRODUCT group ----------
    def get_ref_time(self):
        """
        Returns reference start time of measurements
        """
        ref_time = datetime(2010, 1, 1, 0, 0, 0)
        ref_time += timedelta(seconds=int(self.fid['/PRODUCT/time'][0]))
        return ref_time

    # -------------------------
    def get_delta_time(self):
        """
        Returns offset from the reference start time of measurement
        """
        return self.fid['/PRODUCT/delta_time'][0, :].astype(int)

    # -------------------------
    def __h5_geo_data(self, geo_dsets):
        """
        read gelocation datasets from operational products using HDF5
        """
        res = {}
        if geo_dsets is None:
            geo_dsets = 'latitude,longitude'

        for key in geo_dsets.split(','):
            for grp_name in ['/PRODUCT', '/PRODUCT/SUPPORT_DATA/GEOLOCATIONS']:
                if key in self.fid[grp_name]:
                    res[key] = np.squeeze(
                        self.fid['{}/{}'.format(grp_name, key)])
                    continue

        return res

    def __nc_geo_data(self, geo_dsets):
        """
        read gelocation datasets from science products using netCDF4
        """
        res = {}
        if geo_dsets is None:
            geo_dsets = 'latitude_center,longitude_center'

        for key in geo_dsets.split(','):
            if key in self.fid['/instrument'].variables.keys():
                res[key] = self.fid['/instrument/{}'.format(key)][:]

        return res

    def get_geo_data(self, geo_dsets=None):
        """
        Returns data of selected datasets from the GEOLOCATIONS group

        Parameters
        ----------
        geo_dset  :  string
            Name(s) of datasets, comma separated
            Default:
              * operational: 'latitude,longitude'
              * science: 'latitude_center,longitude_center'

        Returns
        -------
        out   :   dictonary with arrays
           arrays of selected datasets
        """
        if self.science_product:
            return self.__nc_geo_data(geo_dsets)

        return self.__h5_geo_data(geo_dsets)

    # -------------------------
    def __h5_geo_bounds(self, extent, data_sel):
        """
        read bounds of latitude/longitude from operational products using HDF5
        """
        indx = None
        if extent is not None:
            if len(extent) != 4:
                raise ValueError('parameter extent must have 4 elements')

            lats = self.fid['/PRODUCT/latitude'][0, ...]
            lons = self.fid['/PRODUCT/longitude'][0, ...]

            indx = ((lons >= extent[0]) & (lons <= extent[1])
                    & (lats >= extent[2]) & (lats <= extent[3])).nonzero()
            data_sel = np.s_[0,
                             indx[0].min():indx[0].max(),
                             indx[1].min():indx[1].max()]

        gid = self.fid['/PRODUCT/SUPPORT_DATA/GEOLOCATIONS']
        if data_sel is None:
            lat_bounds = gid['latitude_bounds'][0, ...]
            lon_bounds = gid['longitude_bounds'][0, ...]
        else:
            lat_bounds = gid['latitude_bounds'][data_sel + (slice(None),)]
            lon_bounds = gid['longitude_bounds'][data_sel + (slice(None),)]

        return (data_sel, lon_bounds, lat_bounds)

    def __nc_geo_bounds(self, extent, data_sel):
        """
        read bounds of latitude/longitude from science products using netCDF4
        """
        indx = None
        if extent is not None:
            if len(extent) != 4:
                raise ValueError('parameter extent must have 4 elements')

            lats = self.fid['/instrument/latitude_center'][:].reshape(
                self.scanline, self.ground_pixel)
            lons = self.fid['/instrument/longitude_center'][:].reshape(
                self.scanline, self.ground_pixel)

            indx = ((lons >= extent[0]) & (lons <= extent[1])
                    & (lats >= extent[2]) & (lats <= extent[3])).nonzero()
            data_sel = np.s_[indx[0].min():indx[0].max(),
                             indx[1].min():indx[1].max()]

        gid = self.fid['/instrument']
        lat_bounds = gid['latitude_corners'][:].data.reshape(
            self.scanline, self.ground_pixel, 4)
        lon_bounds = gid['longitude_corners'][:].data.reshape(
            self.scanline, self.ground_pixel, 4)
        if data_sel is not None:
            lat_bounds = lat_bounds[data_sel + (slice(None),)]
            lon_bounds = lon_bounds[data_sel + (slice(None),)]

        return (data_sel, lon_bounds, lat_bounds)

    def get_geo_bounds(self, extent=None, data_sel=None):
        """
        Returns bounds of latitude/longitude as a mesh for plotting

        Parameters
        ----------
        extent    :  list
           select data to cover a region with geolocation defined by:
             lon_min, lon_max, lat_min, lat_max and return numpy slice
        data_sel  :  numpy slice
           a 3-dimensional numpy slice: time, scan_line, ground_pixel
           Note 'data_sel' will be overwritten when 'extent' is defined

        Returns
        -------
        data_sel  :   numpy slice
           slice of data which covers geolocation defined by extent. Only
           provided if extent is not None.
        out       :   dictionary
           with numpy arrays for latitude and longitude
        """
        if self.science_product:
            res = self.__nc_geo_bounds(extent, data_sel)
        else:
            res = self.__h5_geo_bounds(extent, data_sel)
        data_sel, lon_bounds, lat_bounds = res

        res = {}
        _sz = lon_bounds.shape
        res['longitude'] = np.empty((_sz[0]+1, _sz[1]+1), dtype=float)
        res['longitude'][:-1, :-1] = lon_bounds[:, :, 0]
        res['longitude'][-1, :-1] = lon_bounds[-1, :, 1]
        res['longitude'][:-1, -1] = lon_bounds[:, -1, 1]
        res['longitude'][-1, -1] = lon_bounds[-1, -1, 2]

        res['latitude'] = np.empty((_sz[0]+1, _sz[1]+1), dtype=float)
        res['latitude'][:-1, :-1] = lat_bounds[:, :, 0]
        res['latitude'][-1, :-1] = lat_bounds[-1, :, 1]
        res['latitude'][:-1, -1] = lat_bounds[:, -1, 1]
        res['latitude'][-1, -1] = lat_bounds[-1, -1, 2]

        if extent is None:
            return res

        return data_sel, res

    # -------------------------
    def __h5_dataset(self, name, data_sel, fill_as_nan):
        """
        read dataset from operational products using HDF5
        """
        fillvalue = float.fromhex('0x1.ep+122')

        if name not in self.fid['/PRODUCT']:
            raise ValueError('dataset {} for found'.format(name))

        dset = self.fid['/PRODUCT/{}'.format(name)]
        if data_sel is None:
            if dset.dtype == np.float32:
                res = dset.astype(float)[0, ...]
            else:
                res = dset[0, ...]
        else:
            if dset.dtype == np.float32:
                res = dset.astype(float)[data_sel]
            else:
                res = dset[data_sel]

        if fill_as_nan and dset.attrs['_FillValue'] == fillvalue:
            res[(res == fillvalue)] = np.nan

        return res

    def __nc_dataset(self, name, data_sel, fill_as_nan):
        """
        read dataset from science products using netCDF4
        """
        if name not in self.fid['/target_product'].variables.keys():
            raise ValueError('dataset {} for found'.format(name))

        dset = self.fid['/target_product/{}'.format(name)]
        res = dset[:].reshape(self.scanline, self.ground_pixel)
        if data_sel is not None:
            res = res[data_sel]

        if fill_as_nan:
            return res.filled(np.nan)

        return res.data

    def get_dataset(self, name, data_sel=None, fill_as_nan=True):
        """
        Read level 2 dataset from PRODUCT group

        Parameters
        ----------
        name   :  string
            name of dataset with level 2 data
        data_sel  :  numpy slice
           a 3-dimensional numpy slice: time, scan_line, ground_pixel
        fill_as_nan :  boolean
            Replace (float) FillValues with Nan's, when True

        Returns
        -------
        out  :  array
        """
        if self.science_product:
            return self.__nc_dataset(name, data_sel, fill_as_nan)

        return self.__h5_dataset(name, data_sel, fill_as_nan)

    # -------------------------
    def __h5_data_as_s5pmsm(self, name, data_sel, fill_as_nan, mol_m2):
        """
        Read dataset from group target_product using HDF5

        Input: operational product

        Return: S5Pmsm object
        """
        if name not in self.fid['/PRODUCT']:
            raise ValueError('dataset {} for found'.format(name))

        dset = self.fid['/PRODUCT/{}'.format(name)]
        msm = S5Pmsm(dset, data_sel=data_sel)
        if '{}_precision'.format(name) in self.fid['/PRODUCT']:
            msm.error = self.get_dataset('{}_precision'.format(name),
                                         data_sel=data_sel)

        if not mol_m2:
            factor = dset.attrs[
                'multiplication_factor_to_convert_to_molecules_percm2']
            msm.value[msm.value != msm.fillvalue] *= factor
            msm.units = 'molecules / cm$^2$'
            if msm.error is not None:
                msm.error[msm.error != msm.fillvalue] *= factor
        else:
            msm.units = 'mol / m$^2$'

        if fill_as_nan:
            msm.fill_as_nan()

        return msm

    def __nc_data_as_s5pmsm(self, name, data_sel, fill_as_nan):
        """
        Read dataset from group PRODUCT using netCDF4

        Input: science product

        Return: S5Pmsm object
        """
        msm = S5Pmsm(self.get_dataset(name, data_sel))
        buff = self.get_dataset('{}_precision'.format(name), data_sel)
        if buff is not None:
            msm.error = buff
        msm.set_fillvalue()
        msm.set_units(self.get_attr('units', name))
        msm.set_long_name(self.get_attr('long_name', name))

        if fill_as_nan:
            msm.fill_as_nan()

        return msm

    def get_data_as_s5pmsm(self, name, data_sel=None, fill_as_nan=True,
                           mol_m2=False):
        """
        Read dataset from group PRODUCT/target_product group

        Parameters
        ----------
        name   :  string
            name of dataset with level 2 data
        data_sel  :  numpy slice
           a 3-dimensional numpy slice: time, scan_line, ground_pixel
        fill_as_nan :  boolean
            Replace (float) FillValues with Nan's, when True
        mol_m2    : boolean
            Leaf units as mol per m^2 or convert units to molecules per cm^2

        Returns
        -------
        out  :  pys5p.S5Pmsm object
        """
        if self.science_product:
            return self.__nc_data_as_s5pmsm(name, data_sel, fill_as_nan)

        return self.__h5_data_as_s5pmsm(name, data_sel, fill_as_nan, mol_m2)
